from time import sleep, time

import pytest

from esteid_certificates import get_certificate

from pyasice import Container, XmlSignature
from pyasice.ocsp import OCSP
from pyasice.tsa import TSA

from ...constants import OCSP_DEMO_URL, TSA_DEMO_URL
from ...exceptions import ActionInProgress
from ..types import SignResult, SignStatusResult


def run_sign_flow(demo_mid_api, certificate, id_number=None, phone_number=None, signed_data=None, timeout=60):
    """Run full sign flow w/ a hash algorithm

    test data from:
    https://github.com/SK-EID/MID/wiki/Test-number-for-automated-testing-in-DEMO

    :param MobileIDService demo_mid_api:
    :param bytes certificate: obtained by authentication, required to validate the signature
    :param str id_number:
    :param str phone_number:
    :param bytes signed_data: content to sign
    :param int timeout: how long to wait for a result. NOTE: if user input is expected, ensure at least 1 minute
    :rtype: SignStatusResult
    """
    if signed_data is None:
        signed_data = b"Test"
    res = demo_mid_api.sign(id_number, phone_number, signed_data)
    assert isinstance(res, SignResult)

    assert res.session_id

    status_res = None  # type: SignStatusResult

    end_time = time() + timeout
    while status_res is None and time() < end_time:
        try:
            status_res = demo_mid_api.sign_status(res.session_id, certificate, res.digest)
        except ActionInProgress:
            sleep(1.0)

    assert isinstance(status_res, SignStatusResult)
    return status_res


@pytest.mark.slow
def test_mobileid_sign_flow_ee(demo_mid_api, MID_DEMO_PHONE_EE_OK, MID_DEMO_PIN_EE_OK):
    """Test full sign flow for EE

    Note: This does a number of real connections to Smart-ID apis so it's marked as a slow test
    """
    file_name = "test.txt"
    data = b"Hello World!"
    mime_type = "text/plain"

    # Authenticate user
    user_cert = demo_mid_api.get_certificate(MID_DEMO_PIN_EE_OK, MID_DEMO_PHONE_EE_OK)

    # Generate a XAdES signature
    xs: XmlSignature = (
        XmlSignature.create().add_document(file_name, data, mime_type).set_certificate(user_cert).update_signed_info()
    )

    # Sign the XAdES structure
    sign_result = run_sign_flow(
        demo_mid_api,
        user_cert,
        id_number=MID_DEMO_PIN_EE_OK,
        phone_number=MID_DEMO_PHONE_EE_OK,
        signed_data=xs.signed_data(),
    )
    xs.set_signature_value(sign_result.signature)

    # prove that all went right
    xs.verify()

    # Get an OCSP status confirmation
    issuer_cert = get_certificate(xs.get_certificate_issuer_common_name())

    ocsp = OCSP(url=OCSP_DEMO_URL)
    ocsp.validate(user_cert, issuer_cert, sign_result.signature)

    # Embed the OCSP response
    xs.set_ocsp_response(ocsp)

    # Get a signature TimeStamp
    tsa = TSA(url=TSA_DEMO_URL)
    tsr = tsa.get_timestamp(xs.get_timestamped_message())
    xs.set_timestamp_response(tsr)

    # Write results to a bdoc file
    with Container() as bdoc_file:
        bdoc_file.add_file(file_name, data, mime_type)
        bdoc_file.add_signature(xs)

        bdoc_file.verify_container()
        bdoc_file.verify_signatures()



from unittest.mock import patch

import pytest

from .. import actions
from ..types import DataFile


@pytest.fixture()
def a_view():
    class AView:
        request = None

        def get_files(self):
            return [DataFile("test.txt", "text/plain", None, 5, b"test!")]

        def get_bdoc_container_file(self):
            return None

    return AView()


@patch.object(actions, "delete_esteid_session")
def test_idcard_prepare_action(_, a_view):
    assert actions.IdCardPrepareAction.do_action(a_view) == {
        "success": False,
        "code": "BAD_CERTIFICATE",
    }


@patch.object(actions, "get_esteid_session")
def test_idcard_finish_action(_, a_view):
    assert actions.IdCardFinishAction.do_action(a_view) == {
        "success": False,
        "code": "BAD_SIGNATURE",
    }



from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

from esteid import settings

from .base import SmartIDService


class TranslatedSmartIDService(SmartIDService):
    """SmartIDService w/ translatable error messages"""

    DISPLAY_TEXT_AUTH = _("Authenticate")
    DISPLAY_TEXT_SIGN = _("Sign")

    @classmethod
    def get_instance(cls):
        cls.configuration_valid()

        return TranslatedSmartIDService(
            rp_uuid=settings.SMART_ID_SERVICE_UUID,
            rp_name=settings.SMART_ID_SERVICE_NAME,
            api_root=settings.SMART_ID_API_ROOT,
        )

    @staticmethod
    def configuration_valid():
        if not (settings.SMART_ID_SERVICE_NAME and settings.SMART_ID_SERVICE_UUID):
            raise ImproperlyConfigured("Both SMART_ID_SERVICE_NAME and SMART_ID_SERVICE_UUID must be set")



import hashlib
import os
import re
from collections import OrderedDict
from datetime import datetime

from django.utils import timezone

from esteid.constants import HASH_ALGORITHMS


ID_CODE_EE_REGEXP = re.compile(r"^[1-6] \d{2} [01]\d [0123]\d \d{4}$", re.VERBOSE)
ID_CODE_LT_REGEXP = ID_CODE_EE_REGEXP
ID_CODE_LV_REGEXP = re.compile(r"^\d{6}-\d{5}$")


def secure_random(n):
    return os.urandom(n)


def generate_hash(algorithm, data):
    """Hash the data with the supplied algorithm

    https://github.com/SK-EID/smart-id-documentation#33-hash-algorithms
    https://github.com/SK-EID/MID#231-supported-hashing-algorithms
    """
    if algorithm.upper() not in HASH_ALGORITHMS:
        raise ValueError(f"Unsupported algorithm {algorithm}")

    hash_method = getattr(hashlib, algorithm.lower())
    digest = hash_method(data).digest()

    return digest


def id_code_ee_is_valid(id_code: str) -> bool:
    """
    Validates Estonian ID code, including checksum.

    https://et.wikipedia.org/wiki/Isikukood
    """
    if isinstance(id_code, str) and bool(re.match(ID_CODE_EE_REGEXP, id_code)):
        step1_factors = "1234567891"
        checksum = sum([int(i) * int(d) for i, d in zip(step1_factors, id_code[:10])]) % 11
        if checksum == 10:
            step2_factors = "3456789123"
            checksum = sum([int(i) * int(d) for i, d in zip(step2_factors, id_code[:10])]) % 11
            if checksum == 10:
                checksum = 0
        if int(id_code[-1]) == checksum:
            return True
    return False


def id_code_lv_is_valid(id_code: str) -> bool:
    """
    Validates Latvian ID code

    Given the input in the following format ABCDEF-XGHIZ,
    Z must equal to (1101-(1*A+6*B+3*C+7*D+9*E+10*F+5*X+8*G+4*H+2*I)) | Mod 11 | Mod 10.
    """
    if isinstance(id_code, str) and bool(re.match(ID_CODE_LV_REGEXP, id_code)):
        id_code = id_code.replace("-", "")
        factors = [1, 6, 3, 7, 9, 10, 5, 8, 4, 2]
        checksum = (1101 - sum([i * int(d) for i, d in zip(factors, id_code[:10])])) % 11 % 10
        if int(id_code[-1]) == checksum:
            return True
    return False


# Lithuanian ID code format is the same as Estonian.
id_code_lt_is_valid = id_code_ee_is_valid


def convert_status(status):
    if isinstance(status, bool):
        return status

    return status.lower() == "ok"


def camel_2_py(the_dict):
    new_dict = {}
    for key, val in the_dict.items():

        if len(re.sub(r"([A-Z])", r" \1", key).split()) == len(key):
            parts = [key.lower()]

        else:
            key = key.replace("ID", "Id").replace("CRL", "Crl")
            parts = re.sub(r"([A-Z])", r" \1", key).split()

        new_dict["_".join([x.lower() for x in parts])] = val

    return new_dict


def convert_time(timestamp):
    if isinstance(timestamp, datetime):
        if timezone.is_naive(timestamp):
            timestamp = timezone.make_aware(timestamp, timezone.utc)

        return timestamp

    return timezone.make_aware(datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ"), timezone.utc)


def get_instance_converter(cls, prepare_kwargs=camel_2_py):
    def _convert_instance(value):
        if isinstance(value, (dict, OrderedDict)):
            return cls(**prepare_kwargs(value))

        return value

    return _convert_instance


def get_typed_list_validator(klass):
    def _get_typed_list_validator(inst, attr, value):
        if not isinstance(value, list):
            raise TypeError("Value MUST be a list")

        if not all(isinstance(x, klass) for x in value):
            raise TypeError("Value MUST be a list of {}".format(klass))

    return _get_typed_list_validator


def get_typed_list_converter(klass):
    converter = get_instance_converter(klass)

    def _get_typed_list_converter(value):
        return [converter(x) for x in value]

    return _get_typed_list_converter


def get_name_from_legacy_common_name(common_name):
    common_name = common_name.replace("\\,", ",")
    common_name = common_name.strip().rstrip(",")

    parts = common_name.split(",")[:-1]
    parts.reverse()

    return " ".join(parts).title()


def get_id_from_legacy_common_name(common_name):
    common_name = common_name.strip().rstrip(",")

    return common_name.split(",")[-1]


def ucs_to_utf8(val):
    return bytes([ord(x) for x in re.sub(r"\\([0-9ABCDEF]{1,2})", lambda x: chr(int(x.group(1), 16)), val)]).decode(
        "utf-8"
    )


def parse_legacy_dn(dn):
    x_client = dn.strip().strip("/").split("/")

    res = {}

    for part in x_client:
        part = ucs_to_utf8(part).split("=")
        res[part[0]] = part[1]

    return res


def parse_rfc_dn(dn):
    dn = ucs_to_utf8(dn).replace("\\,", ",")
    res = {}
    c_key = None

    for part in dn.strip().split(","):
        if "=" in part:
            part = part.split("=")

            c_key = part[0]
            res[c_key] = part[1]

        elif c_key:
            res[c_key] = "{0},{1}".format(res[c_key], part)

    return res



import base64
import pathlib

import pytest

from django.core.files import File

from ..types import DataFile, InterimSessionData


TEST_TXT = "test.txt"
TEST_TXT_CONTENT = b"Hello"


@pytest.fixture()
def test_txt(tmp_path):
    # Can't pass fixtures to parametrize() so we need a well-known file name for a parameterized test.
    # tmp_path is pytest's builtin fixture with a temporary path as a pathlib.Path object
    with open(tmp_path / TEST_TXT, "wb") as f:
        f.write(TEST_TXT_CONTENT)

    yield open(tmp_path / TEST_TXT, "rb")


@pytest.mark.parametrize(
    "value,result",
    [
        pytest.param(
            None,
            TypeError(""),
            id="not a dict",
        ),
        pytest.param(
            {},
            ValueError("Missing required key digest_b64"),
            id="empty dict",
        ),
        pytest.param(
            {"digest_b64": "", "temp_signature_file": "", "temp_container_file": "", "timestamp": 0},
            None,
            id="empty values",
        ),
        pytest.param(
            {"digest_b64": "dGVzdA==", "temp_signature_file": "asdf", "temp_container_file": "asdf", "timestamp": 0},
            None,
            id="non-empty values",
        ),
    ],
)
def test_signing_types_session_data(value, result):
    if isinstance(result, Exception):
        with pytest.raises(type(result)):
            InterimSessionData(value).is_valid()
    else:
        data = InterimSessionData(value)
        assert data.is_valid()
        assert data.digest == base64.b64decode(value["digest_b64"])
        assert data.temp_container_file == value["temp_container_file"]
        assert data.temp_signature_file == value["temp_signature_file"]


@pytest.mark.parametrize(
    "params,result",
    [
        pytest.param(
            {},
            TypeError("required positional argument.+wraps"),
            id="empty params",
        ),
        pytest.param(
            {"wraps": ..., "file_name": TEST_TXT},
            TypeError("Expected a Django File or path to file"),
            id="wraps not a string",
        ),
        pytest.param(
            {"wraps": TEST_TXT},
            TEST_TXT_CONTENT,
            id="wraps a string",
        ),
        pytest.param(
            {"wraps": TEST_TXT, "file_name": "other.name"},
            TEST_TXT_CONTENT,
            id="wraps a string, gets file_name",
        ),
        pytest.param(
            {"wraps": TEST_TXT, "file_name": "other.name", "content": b"test", "mime_type": "text/plain"},
            b"test",
            id="wraps a string, all arguments",
        ),
        pytest.param(
            {"wraps": pathlib.PurePath(TEST_TXT)},
            TEST_TXT_CONTENT,
            id="wraps a PurePath",
        ),
        pytest.param(
            {"wraps": pathlib.PosixPath(TEST_TXT), "file_name": "other.name"},
            TEST_TXT_CONTENT,
            id="wraps a PosixPath",
        ),
        pytest.param(
            {"wraps": pathlib.Path(TEST_TXT), "file_name": "other.name", "content": b"test", "mime_type": "text/plain"},
            b"test",
            id="wraps a Path, all arguments",
        ),
        pytest.param(
            {"wraps": File},
            TEST_TXT_CONTENT,
            id="wraps a File",
        ),
        pytest.param(
            {"wraps": File, "file_name": "other.name"},
            TEST_TXT_CONTENT,
            id="wraps a File, alternative file name",
        ),
        pytest.param(
            {"wraps": File, "file_name": "other.name", "content": b"test", "mime_type": "text/plain"},
            b"test",
            id="wraps a File, all arguments",
        ),
    ],
)
def test_signing_types_data_file(params: dict, result, tmp_path, test_txt):
    if isinstance(result, Exception):
        with pytest.raises(type(result), match=result.args[0]):
            DataFile(**params)

    else:
        # some preparations
        if params["wraps"] is File:
            params["wraps"] = File(test_txt)
        else:
            params["wraps"] = str(tmp_path / TEST_TXT)

        datafile = DataFile(**params)
        assert datafile.wrapped_file == params["wraps"]
        assert datafile.content == params.get("content")
        assert datafile.file_name == (params.get("file_name") or TEST_TXT)
        assert datafile.mime_type == params.get("mime_type", "application/octet-stream")

        data = datafile.read()
        assert data == result



from esteid import context_processors


def test_context_processors_esteid_services(override_esteid_settings):
    test_settings = {
        "ESTEID_DEMO": 1,
        "ID_CARD_ENABLED": 2,
        "MOBILE_ID_ENABLED": 3,
        "MOBILE_ID_TEST_MODE": 4,
        "SMART_ID_ENABLED": 5,
        "SMART_ID_TEST_MODE": 6,
    }

    assert tuple(context_processors.esteid_services().keys()) == tuple(test_settings.keys())

    with override_esteid_settings(**test_settings):
        assert context_processors.esteid_services() == test_settings
